{Version 6.01}
{*********************************************************}
{*                     HTMLVIEW.PAS                      *}
{*                Copyright (c) 1995-8 by                *}
{*                   L. David Baldwin                    *}
{*                 All rights reserved.                  *}
{*********************************************************}

{$i htmlcons.inc}

unit Htmlview;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls, StdCtrls,
  {$ifdef Win32}
    vwPrint,
  {$endif}
  HTMLUn2, Forms, Dialogs, ExtCtrls, ReadHTML, HTMLSubs, Printers, Menus, 
  dsgnintf;  

const
  wm_FormSubmit = wm_User+100;
  wm_DrawBorder = wm_User+101;
  wm_MouseScroll = wm_User+102;

type
  THTMLBorderStyle = (htFocused, htNone, htSingle);
  THotSpotEvent = procedure(Sender: TObject; const SRC: string) of Object;
  THotSpotClickEvent = procedure(Sender: TObject; const SRC: string;
                     var Handled: boolean) of Object;
  TProcessingEvent = procedure(Sender: TObject; ProcessingOn: boolean) of Object;
  TPagePrinted = procedure( Sender: TObject;
                              Canvas : TCanvas ;
                              NumPage, W, H: Integer ;
                              var StopPrinting : Boolean) of Object;
  TPaintPanel = class(TCustomPanel)
  private
    FOnPaint: TNotifyEvent;
    FViewer: TComponent;
    Canvas2: TCanvas;
    Sizing: boolean;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_EraseBkgnd;
    procedure WMLButtonDblClk(var Message: TWMMouse); message WM_LButtonDblClk; 
    procedure Paint; override;
    procedure DoBackground(ACanvas: TCanvas);
    constructor CreateIt(AOwner: TComponent; Viewer: TComponent);
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
  end;

{$ifdef Win32}
  T32ScrollBar = Class(TScrollBar)   {a 32 bit scrollbar}
  private
    FPosition: integer;
    FMin, FMax, FPage: integer;
    procedure SetPosition(Value: integer);
    procedure SetMin(Value: Integer);
    procedure SetMax(Value: Integer);
    procedure CNVScroll(var Message: TWMVScroll); message CN_VSCROLL;
  public
    property Position: integer read FPosition write SetPosition;
    property Min: integer read FMin write SetMin;
    property Max: integer read FMax write SetMax;
    procedure SetParams(APosition, APage, AMin, AMax: Integer);
  end;
{$endif}

  TFileType = (HTMLType, TextType, ImageType);

  THTMLViewer = class(TWinControl)
  private
    { Private declarations }
    DontDraw: boolean;
    FTitle: PString;
    FURL: PString;
    FTarget: PString;
    FBase, FBaseEx: PString;
    FBaseTarget: PString;
    FCurrentFile: PString;
    FNameList: TStringList;
    FCurrentFileType: TFileType;
    FOnHotSpotCovered: THotSpotEvent;
    FOnHotSpotClick: THotSpotClickEvent;
    FOnBitmapRequest: TGetBitmapEvent;
    FOnImageRequest: TGetImageEvent;
    FOnScript: TScriptEvent;
    FOnFormSubmit: TFormSubmitEvent;
    FOnHistoryChange: TNotifyEvent;
    FOnProcessing: TProcessingEvent;
    FHistory, FTitleHistory: TStrings;
    FPositionHistory: TFreeList;
    FHistoryIndex: integer;
    FHistoryMaxCount: integer;
    FFontName: PString;
    FPreFontName: PString;
    FFontColor: TColor;
    FHotSpotColor: TColor;
    FBackGround: TColor;
    FFontSize: integer;
    FProcessing: boolean;
    FAction, FMethod: PString;
    FStringList: TStringList;
    FOldWidth: integer;
    FImageCacheCount: integer;
    FNoSelect: boolean;
    FScrollBars: TScrollStyle;
    FBorderStyle: THTMLBorderStyle;
    FDither: boolean;
    FCaretPos: LongInt;
    sbWidth: integer;
    ScrollWidth: integer;
    MaxVertical: LongInt;
    MouseScrolling: boolean;
    LeftButtonDown: boolean;
    Hiliting: boolean;
    FPrintMarginLeft,
    FPrintMarginRight,
    FPrintMarginTop,
    FPrintMarginBottom: double;
    {$ifdef Ver100}  {Delphi 3}
    FCharset: TFontCharset;
    {$endif}
    FOnPrintHeader, FOnPrintFooter: TPagePrinted;
    FPage: integer;
    HotSpotAction: boolean;
    FMarginHeight, FMarginWidth: integer;
    FServerRoot: string;

    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure ScrollTo(Y: LongInt);
    procedure Scroll(Sender: TObject; ScrollCode: TScrollCode;
           var ScrollPos: Integer);
    procedure Layout;
    procedure SetViewImages(Value: boolean);
    function GetViewImages: boolean;
    procedure SetColor(Value: TColor);
    function GetBase: string;
    procedure SetBase(Value: string);
    function GetBaseTarget: string;
    function GetFURL: string;
    function GetTitle: string;
    function GetCurrentFile: string;
    procedure SetBorderStyle(Value: THTMLBorderStyle);
    function GetPosition: LongInt;
    procedure SetPosition(Value: LongInt);
    function GetScrollPos: integer;
    procedure SetScrollPos(Value: integer);
    function GetScrollBarRange: integer;
    procedure SetHistoryIndex(Value: integer);
    function GetFontName: TFontName;
    procedure SetFontName(Value: TFontName);
    function GetPreFontName: TFontName;
    procedure SetPreFontName(Value: TFontName);
    procedure SetFontSize(Value: integer);
    procedure SetFontColor(Value: TColor);
    procedure SetHotSpotColor(Value: TColor);
    procedure SetOnBitmapRequest(Handler: TGetBitmapEvent);
    procedure SetOnImageRequest(Handler: TGetImageEvent);
    procedure SetOnScript(Handler: TScriptEvent);
    procedure SetOnFormSubmit(Handler: TFormSubmitEvent);
    function GetOurPalette: HPalette;
    procedure SetOurPalette(Value: HPalette);
    procedure SetDither(Value: boolean);
    procedure SetCaretPos(Value: LongInt);
    procedure WMGetDlgCode(var Message: TMessage); message WM_GETDLGCODE;
    procedure BackgroundChange(Sender: TObject);
    procedure SubmitForm(Sender: TObject; Action, Method: string;
                Results: TStringList);
    procedure SetImageCacheCount(Value: integer);
    procedure WMFormSubmit(var Message: TMessage); message WM_FormSubmit;
    procedure WMDrawBorder(var Message: TMessage); message WM_DrawBorder;
    procedure WMMouseScroll(var Message: TMessage); message WM_MouseScroll;
    procedure SetSelLength(Value: LongInt);
    function GetSelLength: LongInt;
    function GetSelText: string;
    procedure SetNoSelect(Value: boolean);
    procedure SetHistoryMaxCount(Value: integer);
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure DrawBorder;
    procedure DoHilite(X, Y: integer);
    procedure SetScrollBars(Value: TScrollStyle);
    procedure SetProcessing(Value: boolean);
    function GetTarget: String;
    {$ifdef Ver100}
    procedure SetCharset(Value: TFontCharset);
    {$endif}
    function GetFormControlList: TList;
    function GetNameList: TStringList;
    procedure SetMarginWidth(Value: integer);
    procedure SetMarginHeight(Value: integer);
    procedure SetServerRoot(Value: string);

  protected
    { Protected declarations }
    SectionList: TSectionList;
    PaintPanel: TPaintPanel;
    PaintBox: TPaintBox;
    {$ifdef Win32}
    VScrollBar: T32ScrollBar;
    {$else}
    VScrollBar: TScrollBar;
    {$endif}
    HScrollBar: TScrollBar;
    Sel1: LongInt;
    procedure DoLogic;
    procedure DoScrollBars;
    procedure SetupAndLogic;
    function GetURL(X, Y: integer; var UrlTarg: TUrlTarget;
             var FormControl: TImageFormControlObj): boolean;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    function GetPalette: HPALETTE; override;
    procedure HTMLPaint(Sender: TObject); virtual;
    procedure HTMLMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer); virtual;
    procedure HTMLMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer); virtual;
    procedure HTMLMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer); virtual;
    procedure HTMLMouseDblClk(X, Y: integer);   
    procedure URLAction; virtual;
    function HotSpotClickHandled: boolean; dynamic;
    procedure LoadFile(const FileName: string; ft: TFileType);
    procedure PaintWindow(DC: HDC); override;
    procedure UpdateImageCache;
    procedure LoadTheStrings(Strings: TStrings; ft: TFileType);

  public
    { Public declarations }
    FrameOwner: TObject;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function HTMLExpandFilename(const Filename: string): string; virtual;
    procedure LoadFromFile(const FileName: string);
    procedure LoadTextFile(const FileName: string);
    procedure LoadImageFile(const FileName: string);
    procedure LoadStrings(Strings: TStrings);
    procedure LoadTextStrings(Strings: TStrings);
    procedure LoadFromBuffer(Buffer: PChar; BufSize: LongInt);
    procedure LoadFromStream(AStream: TStream);
    procedure Print(FromPage, ToPage: integer);
    function PositionTo(Dest: string): boolean;
    function Find(const S: String; MatchCase: boolean): boolean;
    procedure Clear; virtual;
    procedure CopyToClipboard;
    procedure SelectAll;
    procedure ClearHistory;
    procedure Reload;
    procedure BumpHistory(const FileName, Title: string;
                 OldPos: LongInt; ft: TFileType);
    function GetSelTextBuf(Buffer: PChar; BufSize: LongInt): LongInt;
    function NumPrinterPages: integer;

    property DocumentTitle: string read GetTitle;
    property URL: string read GetFURL;
    property Base: string read GetBase write SetBase;
    property BaseTarget: string read GetBaseTarget;
    property Position: LongInt read GetPosition write SetPosition;
    property VScrollBarPosition: integer read GetScrollPos write SetScrollPos;
    property VScrollBarRange: integer read GetScrollBarRange;
    property CurrentFile: string read GetCurrentFile;
    property History: TStrings read FHistory;
    property TitleHistory: TStrings read FTitleHistory;
    property HistoryIndex: integer read FHistoryIndex write SetHistoryIndex;
    property Processing: boolean read FProcessing;
    property SelLength: LongInt read GetSelLength write SetSelLength;
    property SelText: string read GetSelText;
    property Target: string read GetTarget;
    property Palette: HPalette read GetOurPalette write SetOurPalette;
    property Dither: boolean read FDither write SetDither default True;
    property CaretPos: LongInt read FCaretPos write SetCaretPos;
    property FormControlList: TList read GetFormControlList;    
    property NameList: TStringList read GetNameList;

  published
    { Published declarations }
    property OnHotSpotCovered: THotSpotEvent read FOnHotSpotCovered
             write FOnHotSpotCovered;
    property OnHotSpotClick: THotSpotClickEvent read FOnHotSpotClick
             write FOnHotSpotClick;
    property OnBitmapRequest: TGetBitmapEvent read FOnBitmapRequest
             write SetOnBitmapRequest;
    property OnImageRequest: TGetImageEvent read FOnImageRequest
             write SetOnImageRequest;
    property OnScript: TScriptEvent read FOnScript
             write SetOnScript;
    property OnFormSubmit: TFormSubmitEvent read FOnFormSubmit
             write SetOnFormSubmit;
    property OnHistoryChange: TNotifyEvent read FOnHistoryChange
             write FOnHistoryChange;
    property ViewImages: boolean read GetViewImages write SetViewImages;
    property Enabled;
    property TabStop;
    property TabOrder;
    property Align;
    property Name;
    property Tag;
    property PopupMenu;
    property ShowHint;
    property Height default 150;
    property Width default 150;
    property DefBackground: TColor read FBackground write SetColor default clBtnFace;
    property BorderStyle: THTMLBorderStyle read FBorderStyle write SetBorderStyle;
    property Visible;
    property HistoryMaxCount: integer read FHistoryMaxCount write SetHistoryMaxCount;
    property DefFontName: TFontName read GetFontName write SetFontName;
    property DefPreFontName: TFontName read GetPreFontName write SetPreFontName;
    property DefFontSize: integer read FFontSize write SetFontSize default 12;
    property DefFontColor: TColor read FFontColor write SetFontColor
             default clBtnText;
    property DefHotSpotColor: TColor read FHotSpotColor write SetHotSpotColor
             default clBlue;
    property OnProcessing: TProcessingEvent read FOnProcessing write FOnProcessing;
    property ImageCacheCount: integer read FImageCacheCount
                write SetImageCacheCount default 5;
    property NoSelect: boolean read FNoSelect write SetNoSelect;
    property ScrollBars: TScrollStyle read FScrollBars write SetScrollBars default ssBoth;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseDown;
    property OnKeyDown;
    property OnKeyUp;
    property OnKeyPress;
    property OnEnter;
    property PrintMarginLeft: double read FPrintMarginLeft write FPrintMarginLeft;
    property PrintMarginRight: double read FPrintMarginRight write FPrintMarginRight;
    property PrintMarginTop: double read FPrintMarginTop write FPrintMarginTop;
    property PrintMarginBottom: double read FPrintMarginBottom write FPrintMarginBottom;
    property OnPrintHeader: TPagePrinted read FOnPrintHeader write FOnPrintHeader;
    property OnPrintFooter: TPagePrinted read FOnPrintFooter write FOnPrintFooter;
    {$ifdef Ver100}
    property CharSet: TFontCharset read FCharSet write SetCharset;
    {$endif}
    property MarginHeight: integer read FMarginHeight write SetMarginHeight default 5;  
    property MarginWidth: integer read FMarginWidth write SetMarginWidth default 10;  
    property ServerRoot: string read FServerRoot write SetServerRoot;
    end;

  THTMLEditor = class(TComponentEditor)        
    function GetVerbCount: Integer; Override;
    function GetVerb(index: Integer): String; Override;
    procedure ExecuteVerb(index: Integer); Override;
    end;

procedure Register;

implementation

const
  MaxHScroll = 2000;  {max horizontal display in pixels}
  {$ifdef Win32}
  VScale = 1;
  {$else}
  VScale = 16;
  {$endif}

type
  PositionObj = class(TObject)
    Pos: LongInt;
    FileType: TFileType;
    end;

constructor THTMLViewer.Create(AOwner: TComponent);
begin
inherited Create(AOwner);
ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
  csSetCaption, csDoubleClicks];
Height := 150;
Width := 150;
FPrintMarginLeft := 2.0;
FPrintMarginRight := 2.0;
FPrintMarginTop := 2.0;
FPrintMarginBottom := 2.0;
{$ifdef Ver100}  {Delphi 3}
FCharset := DEFAULT_CHARSET;
{$endif}
FMarginHeight := 5;
FMarginWidth := 10;

PaintBox := TPaintBox.Create(Self);
PaintBox.Align := alClient;
InsertControl(PaintBox);

PaintPanel := TPaintPanel.CreateIt(Self, Self);
InsertControl(PaintPanel);
PaintPanel.Top := 1;
PaintPanel.Left := 1;
Cursor := crIBeam;
PaintPanel.Cursor := Cursor;

PaintPanel.OnPaint := HTMLPaint;
PaintPanel.OnMouseDown := HTMLMouseDown;
PaintPanel.OnMouseMove := HTMLMouseMove;
PaintPanel.OnMouseUp := HTMLMouseUp;

{$ifdef Win32}
  VScrollBar := T32ScrollBar.Create(Self);
  VScrollBar.Kind := sbVertical;
  VScrollBar.SmallChange := 16;
{$else}
  VScrollBar := TScrollBar.Create(Self);
  VScrollBar.Kind := sbVertical;
  VScrollBar.SmallChange := 1;
  VScrollBar.OnScroll := Scroll;
{$endif}
VScrollBar.Visible := False;
VScrollBar.TabStop := False;
sbWidth := VScrollBar.Width;
InsertControl(VScrollBar);

HScrollBar := TScrollBar.Create(Self);
HScrollBar.Kind := sbHorizontal;
HScrollBar.SmallChange := 15;
HScrollBar.OnScroll := Scroll;
HScrollBar.Visible := False;
HScrollBar.TabStop := False;
InsertControl(HScrollBar);

FScrollBars := ssBoth;

PaintBox.BringToFront;

SectionList := TSectionList.Create(Self, PaintPanel);
SectionList.OnBackgroundChange := BackgroundChange;
FNameList := TStringList.Create;
FNameList.Sorted := True;
DefBackground := clBtnFace;
DefFontColor := clBtnText;
DefHotSpotColor := clBlue;
DefFontSize := 12;
DefFontName := 'Times New Roman';
DefPreFontName := 'Courier New';
SetImageCacheCount(5);

FBase := NullStr;
FBaseEx := NullStr;
FBaseTarget := NullStr;
FCurrentFile := NullStr;
FTitle := NullStr;
FURL := NullStr;
FTarget := NullStr;

FHistory := TStringList.Create;
FPositionHistory := TFreeList.Create;
FTitleHistory := TStringList.Create;
FDither := True;
end;

destructor ThtmlViewer.Destroy;
begin
SectionList.Free;
FNameList.Free;
DisposeStr(FBase);
DisposeStr(FBaseEx);
DisposeStr(FBaseTarget);
DisposeStr(FCurrentFile);
DisposeStr(FTitle);
DisposeStr(FURL);
DisposeStr(FTarget);
DisposeStr(FFontName);
DisposeStr(FPreFontName);
FHistory.Free;
FPositionHistory.Free;
FTitleHistory.Free;
inherited Destroy;
end;

procedure THtmlViewer.SetupAndLogic;
begin
AssignStr(FTitle, ReadHTML.Title);
if ReadHTML.Base <> '' then         
  AssignStr(FBase, ReadHTML.Base)
else AssignStr(FBase, FBaseEx^);
AssignStr(FBaseTarget, ReadHTML.BaseTarget);
try
  DontDraw := True;
  {Load the background bitmap if any and if ViewImages set}
  SectionList.GetBackgroundBitmap;

DoLogic;

finally
  DontDraw := False;
  end;
end;

procedure ThtmlViewer.LoadFile(const FileName: string; ft: TFileType);
var
  I: integer;
  Dest, FName, OldFile: string;
  SBuffer: string[255];
  OldCursor: TCursor;
begin
with Screen do
  begin
  OldCursor := Cursor;
  Cursor := crHourGlass;
  end;
FName := FileName;
I := Pos('#', FName);
if I > 0 then
  begin
  Dest := copy(FName, I+1, 255);  {positioning information}
  System.Delete(FName, I, 255);
  end
else Dest := '';
try
  SetProcessing(True);
  if not FileExists(FName) then
    Raise(EInOutError.Create('Can''t locate file: '+FName));
  SectionList.Clear;
  UpdateImageCache;
  SectionList.SetFonts(FFontName^, FPreFontName^, FFontSize, FFontColor,
                                    FHotSpotColor, FBackground);
  FNameList.Clear;
  CaretPos := 0;
  Sel1 := -1;
  try
    if ft = HTMLType then
      HtmlParseFile(SectionList, FNameList, FName)
    else if ft = TextType then
      HtmlParseTextFile(SectionList, FNameList, FName)
    else
      begin
      SBuffer := '<img src="'+FName+'">';
      HTMLParseBuffer(SectionList, FNameList, @SBuffer[1], Length(SBuffer));
      end;
  finally
    OldFile := FCurrentFile^;
    AssignStr(FCurrentFile, ExpandFileName(FName));
    FCurrentFileType := ft;
    SetupAndLogic;
    if Dest <> '' then
      PositionTo(Dest)   {change position, if applicable}
    else if FCurrentFile^ <> OldFile then
       begin
       ScrollTo(0);
       HScrollBar.Position := 0;
       end;
    {else if same file leave position alone}
    PaintPanel.Invalidate;
  end;
finally
  Screen.Cursor := OldCursor;
  SetProcessing(False);
  end;
end;

procedure ThtmlViewer.LoadFromFile(const FileName: string);
var
  OldFile, OldTitle: string;
  OldPos: LongInt;
  OldType: TFileType;
begin
if FProcessing then Exit;
if Filename <> '' then
  begin
  OldFile := FCurrentFile^;
  OldTitle := FTitle^;
  OldPos := Position;
  OldType := FCurrentFileType;
  LoadFile(FileName, HTMLType);
  if (OldFile <> FCurrentFile^) or (OldType <> FCurrentFileType) then
    BumpHistory(OldFile, OldTitle, OldPos, OldType);
  end;
end;

{----------------ThtmlViewer.LoadTextFile}
procedure ThtmlViewer.LoadTextFile(const FileName: string);
var
  OldFile, OldTitle: string;
  OldPos: LongInt;
  OldType: TFileType;
begin
if FProcessing then Exit;
if Filename <> '' then
  begin
  OldFile := FCurrentFile^;
  OldTitle := FTitle^;
  OldPos := Position;
  OldType := FCurrentFileType;
  LoadFile(FileName, TextType);
  if (OldFile <> FCurrentFile^) or (OldType <> FCurrentFileType) then
    BumpHistory(OldFile, OldTitle, OldPos, OldType);
  end;
end;

{----------------ThtmlViewer.LoadImageFile}
procedure ThtmlViewer.LoadImageFile(const FileName: string);
var
  OldFile, OldTitle: string;
  OldPos: LongInt;
  OldType: TFileType;

begin
if FProcessing then Exit;
if Filename <> '' then
  begin
  OldFile := FCurrentFile^;
  OldTitle := FTitle^;
  OldPos := Position;
  OldType := FCurrentFileType;
  LoadFile(FileName, ImageType);
  if (OldFile <> FCurrentFile^) or (OldType <> FCurrentFileType) then
    BumpHistory(OldFile, OldTitle, OldPos, OldType);
  end;
end;

procedure THtmlViewer.LoadTheStrings(Strings: TStrings; ft: TFileType);
begin
if FProcessing then Exit;
SetProcessing(True);
try
  SectionList.Clear;
  UpdateImageCache;
  SectionList.SetFonts(FFontName^, FPreFontName^, FFontSize, FFontColor, FHotSpotColor, FBackground);
  FNameList.Clear;
  FCaretPos := 0;
  Sel1 := -1;
  if ft = HTMLType then
    HtmlParseStrings(SectionList, FNameList, Strings)
  else if ft = TextType then
    HtmlParseTextStrings(SectionList, FNameList, Strings);
  SetupAndLogic;
  ScrollTo(0);
  HScrollBar.Position := 0;
  PaintPanel.Invalidate;
  AssignStr(FCurrentFile, '');
finally
  SetProcessing(False);
  end;
end;

{----------------THtmlViewer.LoadStrings}
procedure THtmlViewer.LoadStrings(Strings: TStrings);
begin
LoadTheStrings(Strings, HTMLType);
end;

{----------------THtmlViewer.LoadTextStrings}
procedure THtmlViewer.LoadTextStrings(Strings: TStrings);
begin
LoadTheStrings(Strings, TextType);
end;

{----------------ThtmlViewer.LoadFromBuffer}
procedure ThtmlViewer.LoadFromBuffer(Buffer: PChar; BufSize: LongInt);
begin
if FProcessing then Exit;
SetProcessing(True);
try
  SectionList.Clear;
  UpdateImageCache;
  SectionList.SetFonts(FFontName^, FPreFontName^, FFontSize, FFontColor,
                                    FHotSpotColor, FBackground);
  FNameList.Clear;
  CaretPos := 0;
  Sel1 := -1;
  HtmlParseBuffer(SectionList, FNameList, Buffer, BufSize);
  SetupAndLogic;
  ScrollTo(0);
  HScrollBar.Position := 0;
  PaintPanel.Invalidate;
  AssignStr(FCurrentFile, '');
finally
  SetProcessing(False);
  end;
end;

{----------------ThtmlViewer.LoadFromStream}
procedure ThtmlViewer.LoadFromStream(AStream: TStream);
begin
if FProcessing then Exit;
SetProcessing(True);
try
  SectionList.Clear;
  UpdateImageCache;
  SectionList.SetFonts(FFontName^, FPreFontName^, FFontSize, FFontColor,
                                    FHotSpotColor, FBackground);
  FNameList.Clear;
  CaretPos := 0;
  Sel1 := -1;
  HtmlParseStream(SectionList, FNameList, AStream);
  SetupAndLogic;
  ScrollTo(0);
  HScrollBar.Position := 0;
  PaintPanel.Invalidate;
  AssignStr(FCurrentFile, '');
finally
  SetProcessing(False);
  end;
end;

{----------------ThtmlViewer.DoScrollBars}
procedure ThtmlViewer.DoScrollBars;
var
  VBar, HBar: boolean;
  Wid, HWidth, WFactor: integer;
{$ifdef Win32}
  ScrollInfo :TScrollInfo;
{$endif}

begin
VBar := False;
ScrollWidth := IntMin(ScrollWidth, MaxHScroll);    
if FBorderStyle = htNone then
  begin
  WFactor := 0;
  PaintPanel.Top := 0;
  PaintPanel.Left := 0;
  end
else
  begin
  WFactor := 1;
  PaintPanel.Top := 1;
  PaintPanel.Left := 1;
  end;
if FScrollBars in [ssBoth, ssVertical] then
  begin  {assume a vertical scrollbar}
  VBar := (MaxVertical >= Height-2) or
          ((FScrollBars in [ssBoth, ssHorizontal]) and
           (MaxVertical >= Height-2-sbWidth) and
           (ScrollWidth+2*FMarginWidth > Width-sbWidth));
  HBar := (FScrollBars in [ssBoth, ssHorizontal]) and
          ((ScrollWidth+2*FMarginWidth > Width) or
           (VBar and (ScrollWidth+2*FMarginWidth > Width-sbWidth)));
  end
else
  begin  {there is no vertical scrollbar}
  HBar := (FScrollBars in [ssBoth, ssHorizontal]) and
          (ScrollWidth+2*FMarginWidth > Width);
  end;
if VBar then
  Wid := Width - sbWidth
else
  Wid := Width;
PaintPanel.Width := Wid - 2*WFactor;
if HBar then
  PaintPanel.Height := Height - 2*WFactor - sbWidth
else
  PaintPanel.Height := Height - 2*WFactor;
HWidth := IntMax(ScrollWidth+2*FMarginWidth, Wid-2*WFactor);
HScrollBar.Visible := HBar;
HScrollBar.LargeChange := IntMax(1, Wid - 20);
HScrollBar.SetBounds(WFactor, Height-sbWidth-WFactor, Wid -WFactor, sbWidth);
VScrollBar.SetBounds(Width-sbWidth-WFactor, WFactor, sbWidth, PaintPanel.Height);
VScrollBar.LargeChange := PaintPanel.Height div VScale - VScrollBar.SmallChange;
VScrollBar.Visible := VBar;

{$ifndef Win32}
VScrollBar.SetParams(VScrollBar.Position, 0,
          IntMax(0, (MaxVertical-PaintPanel.Height) div VScale + 2));
HScrollBar.Max := IntMax(0, HWidth  - Wid);
{$else}
HScrollBar.Max := IntMax(0, HWidth);
VScrollBar.SetParams(VScrollBar.Position, PaintPanel.Height+1, 0, MaxVertical);
ScrollInfo.cbSize := SizeOf(ScrollInfo);
ScrollInfo.fMask := SIF_PAGE;
ScrollInfo.nPage := Wid;
SetScrollInfo(HScrollBar.Handle,SB_CTL,ScrollInfo,TRUE);
{$endif}
end;

{----------------ThtmlViewer.DoLogic}
procedure ThtmlViewer.DoLogic;
var
  Curs: LongInt;
  Wid: integer;
begin
ScrollWidth := 0;
Curs := 0;
try

  DontDraw := True;
  if FScrollBars in [ssBoth, ssVertical] then
    begin  {assume a vertical scrollbar}
    Wid := Width - sbWidth;
    MaxVertical := SectionList.DoLogic(PaintPanel.Canvas, FMarginHeight,
            Wid-2*FMarginWidth, ScrollWidth, Curs) + 2*FMarginHeight;     
    DoScrollBars;
    end
  else
    begin  {there is no vertical scrollbar}
    Wid := Width;
    MaxVertical := SectionList.DoLogic(PaintPanel.Canvas, FMarginHeight,
            Wid-2*FMarginWidth, ScrollWidth, Curs) + FMarginHeight;
    DoScrollBars;
    end;
finally

  DontDraw := False;
  end;
end;

procedure ThtmlViewer.HTMLPaint(Sender: TObject);
var
  ARect: TRect;
begin
if not DontDraw then
  begin
  ARect := Rect(FMarginWidth, 1, PaintPanel.Width, PaintPanel.Height);
  SectionList.Draw(PaintPanel.Canvas2, ARect, MaxHScroll,
                        FMarginWidth - HScrollBar.Position, FMarginHeight);
  end;
end;

procedure ThtmlViewer.WMSize(var Message: TWMSize);
begin
inherited;
if (Width <> FOldWidth) and not FProcessing then   
  begin
  PaintPanel.Sizing := True;
  Layout;
  FOldWidth := Width;
  end
else
  DoScrollBars;
if MaxVertical < PaintPanel.Height then
  Position := 0
else ScrollTo(VScrollBar.Position * LongInt(VScale));   {keep aligned to limits}
with HScrollBar do
  {$ifdef Win32}
  Position := IntMin(Position, Max - PaintPanel.Width);
  {$else}
  Position := IntMin(Position, Max);
  {$endif}
end;

{$ifdef Win32}
procedure ThtmlViewer.Scroll(Sender: TObject; ScrollCode: TScrollCode;
       var ScrollPos: Integer);
{only the 32 bit horizontal scrollbar comes here}
begin
SetFocus;
ScrollPos := IntMin(ScrollPos, HScrollBar.Max - PaintPanel.Width);
PaintPanel.Invalidate;
end;

{$else}
procedure ThtmlViewer.Scroll(Sender: TObject; ScrollCode: TScrollCode;
       var ScrollPos: Integer);
begin
SetFocus;
if Sender = VScrollBar then
  SectionList.SetYOffset(ScrollPos * LongInt(VScale));
PaintPanel.Invalidate;
end;
{$endif}

procedure ThtmlViewer.ScrollTo(Y: LongInt);
{$ifndef Win32}
var
  Tmp: LongInt;
{$endif}
begin
{$ifdef Win32}
  Y := IntMin(Y, MaxVertical - PaintPanel.Height);
  Y := IntMax(Y, 0);
  VScrollBar.Position := Y;
{$else}
  Tmp := VScrollBar.Max * LongInt(VScale);
  if Y > Tmp then
    Y := Tmp;
  if Y < 0 then Y := 0;
  VScrollBar.Position := Y div VScale;
{$endif}
SectionList.SetYOffset(Y);
Invalidate;
end;

procedure ThtmlViewer.Layout;
var
  OldPos: LongInt;
begin
if FProcessing then Exit;
SetProcessing(True);
try
  OldPos := Position;
  DoLogic;
  Position := OldPos;   {return to old position after width change}
finally
  SetProcessing(False);
  end;
end;

function ThtmlViewer.HotSpotClickHandled: boolean;
var
  Handled: boolean;
begin
Handled := False;
if Assigned(FOnHotSpotClick) then
  FOnHotSpotClick(Self, URL, Handled);
Result := Handled;
end;

procedure ThtmlViewer.URLAction;
var
  S, Dest: string;
  Ext: string[5];
  I: integer;
  OldPos: LongInt;

begin
if not HotSpotClickHandled then
  begin
  OldPos := Position;
  S := URL;
  I := Pos('#', S);  {# indicates a position within the document}
  if I = 1 then
    begin
    if PositionTo(S) then    {no filename with this one}
      BumpHistory(FCurrentFile^, FTitle^, OldPos, FCurrentFileType);
    end
  else
    begin
    if I >= 1 then
      begin
      Dest := System.Copy(S, I, 255);  {local destination}
      S := System.Copy(S, 1, I-1);     {the file name}
      end
    else
      Dest := '';    {no local destination}
    S := HTMLExpandFileName(S);
    Ext := Uppercase(ExtractFileExt(S));
{$ifdef Win32}
    if (Ext = '.HTM') or (Ext = '.HTML')  then
{$else}
    if Ext = '.HTM' then
{$endif}
      begin              {an html file}
      if S <> FCurrentFile^ then
        LoadFromFile(S + Dest)
      else
        if PositionTo(Dest) then   {file already loaded, change position}
          BumpHistory(FCurrentFile^, FTitle^, OldPos, HTMLType);
      end
    else if (Ext = '.BMP') or (Ext = '.GIF') or (Ext = '.JPG') or (Ext = '.JPEG')
                or (Ext = '.PNG') then
      LoadImageFile(S);
    end;
  end;
end;

procedure ThtmlViewer.HTMLMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  XR, CaretHt: integer;
  YR: LongInt;
  Cell1: TObject;
  WasFocused: boolean;
begin
inherited MouseDown(Button, Shift, X, Y);

WasFocused := Focused;
HiLiting := WasFocused;
SetFocus;
HotSpotAction := False;    
if (Button <> mbLeft) then Exit;
LeftButtonDown := True;          
if not WasFocused then Exit;
with SectionList do
  begin
  Sel1 := FindCursor(PaintPanel.Canvas, X, Y+YOff-FMarginHeight, XR, YR, CaretHt, Cell1); 
  if Sel1 > -1 then
    begin
    if SelB <> SelE then
      InvalidateRect(PaintPanel.Handle, Nil, True);
    SelB := Sel1;
    SelE := Sel1;
    CaretPos := Sel1;
    end;
  end;
end;

procedure ThtmlViewer.HTMLMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  UrlTarget : TUrlTarget;
  Url, Target: string;
  FormControl: TImageFormControlObj;
begin
Inherited MouseMove(Shift,X,Y);

UrlTarget := Nil;
URL := '';
if GetURL(X, Y, UrlTarget, FormControl) then  
  begin
  PaintPanel.Cursor := HandCursor;
  if not Assigned(FormControl) then
    begin
    Url := UrlTarget.Url^;
    Target := UrlTarget.Target^;
    UrlTarget.Free;
    end;
  end
else PaintPanel.Cursor := Cursor;
if URL <> FURL^ then
  begin
  AssignStr(FURL, URL);
  AssignStr(FTarget, Target);
  if Assigned(FOnHotSpotCovered) then FOnHotSpotCovered(Self, URL);
  end;
if (ssLeft in Shift) and not MouseScrolling and ((Y <= 0) or (Y > Self.Height)) then
  begin
  MouseScrolling := True;
  PostMessage(Handle, wm_MouseScroll, 0, 0);
  end;
if (ssLeft in Shift) and not FNoSelect then
  DoHilite(X, Y);
inherited MouseMove(Shift, X, Y);
end;

procedure ThtmlViewer.HTMLMouseUp(Sender: TObject; Button: TMouseButton;
            Shift: TShiftState; X, Y: Integer);
var
  UrlTarget: TUrlTarget;
  FormControl: TImageFormControlObj;
begin
inherited MouseUp(Button, Shift, X, Y);

if Button <> mbLeft then Exit;
DoHilite(X, Y);
Hiliting := False;
MouseScrolling := False;
if LeftButtonDown and (SectionList.SelE <= SectionList.SelB)  
       and GetURL(X, Y, UrlTarget, FormControl) then  
  if Assigned(FormControl) then
    FormControl.ImageClick
  else
    begin
    AssignStr(FURL, UrlTarget.URL^);
    AssignStr(FTarget, UrlTarget.Target^);
    UrlTarget.Free;
    HotSpotAction := True;   {prevent double click action} 
    URLAction;
    end;
LeftButtonDown := False;  
end;

{----------------ThtmlViewer.HTMLMouseDblClk}
procedure ThtmlViewer.HTMLMouseDblClk(X, Y: integer);
const
  AlphNum = ['a'..'z', 'A'..'Z', '0'..'9', #192..#214, #216..#246, #248..#255];   
var
  st, en: LongInt;
  XR, X1, CaretHt: integer;
  YR, Y1: LongInt;
  Cell1: TObject;
  Obj: TObject;
  Ch: char;

  function GetCh(Pos: integer): char;
  var
    Ch: char;
    Obj1: TObject;
  begin
  Result := ' ';
  if not SectionList.GetChAtPos(Pos, Ch, Obj1) or (Obj1 <> Obj) then Exit;
  Result := Ch;
  end;
                         
begin
if FNoSelect or FProcessing or HotSpotAction then Exit;
with SectionList do
  begin
  CaretPos := FindCursor(PaintPanel.Canvas, X, Y+YOff-FMarginHeight, XR, YR, CaretHt, Cell1); 
  CursorToXY(PaintPanel.Canvas, CaretPos, X1, Y1);
  if Y1-FMarginHeight <> YR then Exit;  {cursor is past end of row}
  en := CaretPos;
  st := en-1;
  if not GetChAtPos(en, Ch, Obj) then Exit;
  if not (Ch in AlphNum) then Exit;
  Inc(en);
  while GetCh(en) in AlphNum do
    Inc(en);
  while (st >= 0) and (GetCh(st) in AlphNum) do
    Dec(st);
  SelB := st+1;
  SelE := en;
  InvalidateRect(PaintPanel.Handle, Nil, True);
  end;
end;

procedure ThtmlViewer.DoHilite(X, Y: integer);
var
  Curs, YR, YWin: LongInt;
  CursCell: TObject;
  XR, CaretHt: integer;
begin
if Hiliting and (Sel1 >= 0) then
  with SectionList do
    begin
    CursCell := Nil;
    YWin := IntMin(IntMax(0, Y), Height);
    Curs := FindCursor(PaintPanel.Canvas, X, YWin+YOff-FMarginHeight, XR, YR, CaretHt, CursCell); 
    if (Curs <> CaretPos) and (Curs >= 0) and not FNoSelect then
      begin
      if Curs > Sel1 then
        begin
        SelE := Curs;
        SelB := Sel1;
        end
      else
        begin
        SelB := Curs;
        SelE := Sel1;
        end;
      InvalidateRect(PaintPanel.Handle, Nil, True);
      end;
    CaretPos := Curs;
    end;
end;

procedure ThtmlViewer.WMMouseScroll(var Message: TMessage);
const
  Ticks: LongInt = 0;
var
  Pos: integer;
  Pt: TPoint;
begin
GetCursorPos(Pt);
Ticks := 0;
with VScrollBar do
  begin
  Pt := PaintPanel.ScreenToClient(Pt);
  while MouseScrolling and ((Pt.Y <= 0) or (Pt.Y > Self.Height)) do
    begin
    if GetTickCount > Ticks +100 then
      begin
      Ticks := GetTickCount;
      if Pt.Y < -15 then
        Pos := Position - SmallChange * 8
      else if Pt.Y <= 0 then
        Pos := Position - SmallChange
      else if Pt.Y > Self.Height+15 then
        Pos := Position + SmallChange * 8
      else
        Pos := Position + SmallChange;
{$ifdef Win32}
      Pos := IntMax(0, IntMin(Pos, MaxVertical - PaintPanel.Height));
      SectionList.SetYOffset(Pos * LongInt(VScale));
      SetPosition(Pos);
{$else}
      Pos := IntMax(0, IntMin(Pos, Max));
      Position := Pos;
      Self.Scroll(VScrollBar, scLineUp, Pos);
{$endif}
      DoHilite(Pt.X, Pt.Y);
      PaintPanel.Invalidate;
      GetCursorPos(Pt);
      Pt := PaintPanel.ScreenToClient(Pt);
      end;
    Application.ProcessMessages;
    end;
  end;
MouseScrolling := False;
end;

function ThtmlViewer.PositionTo(Dest: string): boolean;
var
  I: integer;
begin
Result := False;
If Dest = '' then Exit;
if Dest[1] = '#' then
  System.Delete(Dest, 1, 1);
I := FNameList.IndexOf(UpperCase(Dest));
if I > -1 then
  begin
  ScrollTo(TSectionBase(FNameList.Objects[I]).YValue);
  HScrollBar.Position := 0;
  Result := True;
  end;
end;

function ThtmlViewer.GetURL(X, Y: integer; var UrlTarg: TUrlTarget;
          var FormControl: TImageFormControlObj): boolean;
begin
Result := SectionList.GetURL(PaintPanel.Canvas, X, Y+SectionList.YOff-FMarginHeight,   
          UrlTarg, FormControl);
end;

procedure THTMLViewer.SetViewImages(Value: boolean);
var
  OldPos: LongInt;
  OldCursor: TCursor;
begin
if (Value <> SectionList.ShowImages) and not FProcessing then
  begin
  OldCursor := Screen.Cursor;
  try
    Screen.Cursor := crHourGlass;
    SetProcessing(True);
    SectionList.ShowImages := Value;
    if SectionList.Count > 0 then
      begin
      SectionList.GetBackgroundBitmap;    {load any background bitmap}
      OldPos := Position;
      DoLogic;
      Position := OldPos;
      Invalidate;
      end;
  finally
    Screen.Cursor := OldCursor;
    SetProcessing(False);
    end;
  end;
end;

function THTMLViewer.GetBase: string;
begin
Result := FBase^;
end;

procedure THTMLViewer.SetBase(Value: string);
begin
AssignStr(FBase, Value);
AssignStr(FBaseEx, Value);   
end;

function THTMLViewer.GetBaseTarget: string;
begin
Result := FBaseTarget^;
end;

function THTMLViewer.GetTitle: string;
begin
Result := FTitle^;
end;

function THTMLViewer.GetCurrentFile: string;
begin
Result := FCurrentFile^;
end;

function THTMLViewer.GetFURL: string;
begin
Result := FURL^;
end;

function THTMLViewer.GetTarget: string;
begin
Result := FTarget^;
end;

function THTMLViewer.GetViewImages: boolean;
begin
Result := SectionList.ShowImages;
end;

procedure THTMLViewer.SetColor(Value: TColor);
begin
if FProcessing then Exit;
FBackground := Value;
SectionList.Background:= Value;
PaintPanel.Color := Value;
Invalidate;
end;

procedure THTMLViewer.SetBorderStyle(Value: THTMLBorderStyle);
begin
if Value <> FBorderStyle then
  begin
  FBorderStyle := Value;
  DrawBorder;
  end;
end;

procedure ThtmlViewer.KeyDown(var Key: Word; Shift: TShiftState);
var
  Pos: integer;
  OrigPos:    integer;
  TheChange: LongInt;
begin
inherited KeyDown(Key, Shift);
with VScrollBar do
  if Key in [VK_PRIOR, VK_NEXT, VK_UP, VK_DOWN, VK_HOME, VK_END] then
    begin
    Pos := Position;
    OrigPos := Pos;
    case Key of
      VK_PRIOR : Dec(Pos, LargeChange);
      VK_NEXT  : Inc(Pos, LargeChange);
      VK_UP    : Dec(Pos, SmallChange);
      VK_DOWN  : Inc(Pos, SmallChange);
      VK_Home  : Pos := 0;
      VK_End   : Pos := MaxVertical div VScale;
      end;
    if Pos < 0 then Pos := 0;
    {$ifdef Win32}
    Pos := IntMax(0, IntMin(Pos, MaxVertical - PaintPanel.Height));
    {$else}
    Pos := IntMin(Pos, Max);
    {$endif}

    Position := Pos;
    SectionList.SetYOffset(Pos * LongInt(VScale));
    TheChange := OrigPos-Pos;
    if abs(TheChange) = SmallChange then begin
       ScrollWindow(PaintPanel.Handle, 0, TheChange*VScale, NIL, NIL);
       PaintPanel.Update;
    end else
       PaintPanel.Invalidate;
    end;
with HScrollBar do
  if Key in [VK_LEFT, VK_RIGHT] then
    begin
    Pos := Position;
    case Key of
      VK_LEFT  : Dec(Pos, SmallChange);
      VK_RIGHT : Inc(Pos, SmallChange);
      end;
    if Pos < 0 then Pos := 0;
    {$ifdef Win32}
    Pos := IntMin(Pos, Max - PaintPanel.Width);
    {$else}
    Pos := IntMin(Pos, Max);
    {$endif}
    Position := Pos;
    PaintPanel.Invalidate;
    end;
end;

procedure ThtmlViewer.WMGetDlgCode(var Message: TMessage);
begin
Message.Result := DLGC_WantArrows;  {else don't get the arrow keys}
end;

function ThtmlViewer.GetPosition: LongInt;
var
  Index: integer;
  TopPos, Pos: LongInt;
  S: TSectionBase;
begin
Pos := LongInt(VScrollBar.Position) * VScale;
S:= SectionList.FindSectionAtPosition(Pos, TopPos, Index);
if Assigned(S) then
  Result := LongInt(Index) shl 16 + ((Pos - TopPos) and $FFFF)
else Result := 0;
{Hiword is section #, Loword is displacement from top of section}
end;

procedure ThtmlViewer.SetPosition(Value: LongInt);
var
  TopPos: LongInt;
begin
if (HiWord(Value) >= 0) and (Hiword(Value) < SectionList.Count) then
  begin
  TopPos := SectionList.FindPositionByIndex(HiWord(Value));
  ScrollTo(TopPos + LoWord(Value));
  end;
end;

function ThtmlViewer.GetScrollPos: integer;
begin
Result := VScrollBar.Position;
end;

procedure ThtmlViewer.SetScrollPos(Value: integer);
begin
    if Value < 0 then Value := 0;
    {$ifdef Win32}
    Value := IntMin(Value, MaxVertical - PaintPanel.Height);
    {$else}
    Value := IntMin(Value, VScrollBar.Max);
    {$endif}
if Value <> GetScrollPos then
  ScrollTo(LongInt(Value) * VScale);
end;

function ThtmlViewer.GetScrollBarRange: integer;
begin
Result := VScrollBar.Max;
end;

function ThtmlViewer.GetPalette: HPALETTE;
begin
if ThePalette <> 0 then
  Result := ThePalette
else Result := inherited GetPalette;
Invalidate;
end;

function ThtmlViewer.HTMLExpandFilename(const Filename: string): string;
begin
Result := HTMLServerToDos(Trim(Filename), FServerRoot);   
if Pos('\', Result) = 1 then
  Result := ExpandFilename(Result)
else if Pos(':', Result)<> 2 then
  if CompareText(FBase^, 'DosPath') = 0 then  {let Dos find the path}
  else if FBase^ <> '' then
    Result := ExpandFilename(HTMLToDos(FBase^) + Result)   
  else
    Result := ExpandFilename(ExtractFilePath(FCurrentFile^) + Result);  
end;

{----------------ThtmlViewer.BumpHistory}
procedure ThtmlViewer.BumpHistory(const FileName, Title: string;
              OldPos: LongInt; ft: TFileType);
var
  I: integer;
  PO: PositionObj;
begin
if (FHistoryMaxCount > 0) and  (FCurrentFile^ <> '') and
         ((FileName <> FCurrentFile^) or (FCurrentFileType <> ft)
         or (OldPos <> Position)) then
  with FHistory do
    begin
    if (Count > 0) and (Filename <> '') then
      begin
      Strings[FHistoryIndex] := Filename;
      with PositionObj(FPositionHistory[FHistoryIndex]) do
        begin
        Pos := OldPos;
        FileType := ft;
        end;
      FTitleHistory[FHistoryIndex] := Title;
      for I := 0 to FHistoryIndex-1 do
        begin
        Delete(0);
        FTitleHistory.Delete(0);
        PositionObj(FPositionHistory[0]).Free;
        FPositionHistory.Delete(0);
        end;
      end;
    FHistoryIndex := 0;
    Insert(0, FCurrentFile^);
    PO := PositionObj.Create;
    PO.Pos := Position;
    PO.FileType := FCurrentFileType;
    FPositionHistory.Insert(0, PO);
    FTitleHistory.Insert(0, FTitle^);
    if Count > FHistoryMaxCount then
      begin
      Delete(FHistoryMaxCount);
      FTitleHistory.Delete(FHistoryMaxCount);
      PositionObj(FPositionHistory[FHistoryMaxCount]).Free;
      FPositionHistory.Delete(FHistoryMaxCount);
      end;
    if Assigned(FOnHistoryChange) then FOnHistoryChange(Self);
    end;
end;

procedure ThtmlViewer.SetHistoryIndex(Value: integer);
begin
with FHistory do
  if (Value <> FHistoryIndex) and (Value >= 0) and (Value < Count)
            and not FProcessing then
    begin
    if FCurrentFile^ <> '' then
      begin
      Strings[FHistoryIndex] := FCurrentFile^;
      with PositionObj(FPositionHistory[FHistoryIndex]) do
        begin
        Pos := Position;
        FileType := FCurrentFileType;
        end;
      FTitleHistory[FHistoryIndex] := FTitle^;
      end;
    with PositionObj(FPositionHistory[Value]) do
      begin
      if (FCurrentFile^ <> Strings[Value]) or (FCurrentFileType <> FileType) then
        Self.LoadFile(Strings[Value], FileType);
      Position := Pos;
      end;
    FHistoryIndex := Value;
    if Assigned(FOnHistoryChange) then FOnHistoryChange(Self);
    end;
end;

procedure ThtmlViewer.SetHistoryMaxCount(Value: integer);
begin
if (Value = FHistoryMaxCount) or (Value < 0) then Exit;
if Value < FHistoryMaxCount then
  ClearHistory;
FHistoryMaxCount := Value;
end;

procedure ThtmlViewer.ClearHistory;
var
  CountWas: integer;
begin
CountWas := FHistory.Count;
FHistory.Clear;
FTitleHistory.Clear;
FPositionHistory.Clear;
FHistoryIndex := 0;
if (CountWas > 0) and Assigned(FOnHistoryChange) then
  FOnHistoryChange(Self);
end;

function ThtmlViewer.GetFontName: TFontName;
begin
Result := FFontName^;
end;

procedure ThtmlViewer.SetFontName(Value: TFontName);
begin
if  CompareText(Value, SectionList.FontName) <> 0 then
  begin
  DisposeStr(FFontName);
  FFontName := NewStr(Value);
  SectionList.FontName := Value;
  SectionList.UpdateFonts;
  if SectionList.Count > 0 then
    Layout;
  Invalidate;
  end;
end;

function ThtmlViewer.GetPreFontName: TFontName;
begin
Result := FPreFontName^;
end;

procedure ThtmlViewer.SetPreFontName(Value: TFontName);
begin
if  CompareText(Value, SectionList.PreFontName) <> 0 then
  begin
  DisposeStr(FPreFontName);
  FPreFontName := NewStr(Value);
  SectionList.PreFontName := Value;
  SectionList.UpdateFonts;
  if SectionList.Count > 0 then
    Layout;
  Invalidate;
  end;
end;

procedure ThtmlViewer.SetFontSize(Value: integer);
begin
Value := IntMax(Value, 6);  {minimum value of 6 pts}
FFontSize := Value;
SectionList.FontSize := Value;
SectionList.UpdateFonts;
if SectionList.Count > 0 then
  Layout;
Invalidate;
end;

{$ifdef Ver100}
procedure ThtmlViewer.SetCharset(Value: TFontCharset);
begin
FCharset := Value;
SectionList.Charset := Value;
SectionList.UpdateFonts;
if SectionList.Count > 0 then
  Layout;
Invalidate;
end;
{$endif}

function ThtmlViewer.GetFormControlList: TList;
begin
Result := SectionList.FormControlList;
end;

function ThtmlViewer.GetNameList: TStringList;
begin
Result := FNameList;
end;

procedure ThtmlViewer.SetFontColor(Value: TColor);
begin
FFontColor := Value;
SectionList.FontColor := Value;
SectionList.UpdateFonts;
Invalidate;
end;

procedure ThtmlViewer.SetHotSpotColor(Value: TColor);
begin
FHotSpotColor := Value;
SectionList.HotSpotColor := Value;
SectionList.UpdateFonts;
Invalidate;
end;

{$ifdef Windows}  {the Delphi 1 version}
procedure ThtmlViewer.Print(FromPage, ToPage: integer);
var
  ARect: TRect;
  PrintList: TSectionList;
  P1, P2, P3, W, H, Index, Dummy: integer;
  HTop, HBot: LongInt;
  LineTop, Curs: LongInt;
  Done: boolean;
  DC : HDC;
  SaveFont: TFont;
  SaveSize: integer;

  { added to query the printer for printable area }
  QEsc: integer;

  UpperLeftPagePoint, { these will contain Top/Left and Bottom/Right
                        unprintable area}
  LowerRightPagePoint: TPoint;

  { added for "local" coordinate system }
  MLeft: integer;

  MLeftPrn: integer;
  MRightPrn: integer;
  MTopPrn: integer;
  MBottomPrn: integer;
  TopPixels, TopPixelsPrn, HPrn, WPrn: integer;
  hrgnClip: THandle;
  savedFont : TFont ;
  savedPen : TPen ;
  savedBrush : TBrush ;
  Align, ScaledPgHt, ScaledPgWid: integer;

begin
FPage := 0;
if FProcessing or (SectionList.Count = 0) then Exit;     
PrintList := TSectionList.CreateCopy(SectionList);
PrintList.SetYOffset(0);
try
  savedFont := TFont.Create ;
  savedPen := TPen.Create ;
  savedBrush := TBrush.Create ;
  try
    PrintList.Printing := True;
    PrintList.SetBackground(clWhite);
    FPage := 1;
    try
      with Printer, Canvas do
        begin
        hRgnClip := 0;
        if (Printer.Title = '') and (DocumentTitle <> '') then   
            Printer.Title := DocumentTitle ;
        BeginDoc;
        DC := Canvas.Handle;
        SaveFont := TFont.Create;
        SaveFont.Assign(Canvas.Font);
        P3 := GetDeviceCaps(DC, LOGPIXELSY);
        P2 := Screen.PixelsPerInch;
        Canvas.Font.PixelsPerInch := P2;  {use screen pixels for ThtmlViewer}
        SetMapMode(DC, mm_AnIsotropic);
        P1 := GetDeviceCaps(DC, LOGPIXELSX);
        SetWindowExtEx(DC, P2, P2, Nil);
        SetViewPortExtEx(DC, P1,P3, Nil);

        { calculate the amount of space that is non-printable }

        { get PHYSICAL page width }
        QEsc := GetPhysPageSize;
        if Escape(Printer.Handle, QueryEscSupport, sizeof(Integer),
                                  @QEsc, NIL)>0 then
          Escape(Printer.Handle, GetPhysPageSize, 0, NIL, @LowerRightPagePoint)
        else { command NOT supported !}
          with LowerRightPagePoint do
          begin
            x := -1; y := -1; { signal that these values are not good at all...}
          end;

        { now compute a complete unprintable area rectangle (composed of 2*width,
          2*height) in pixels...}
        with LowerRightPagePoint do
        begin
          y := y - GetDeviceCaps(DC, VertRes);
          x := x - GetDeviceCaps(DC, HorzRes);
        end;

        { get upper left physical offset for the printer... ->
          printable area <> paper size }
        QEsc := GetPrintingOffset;
        if Escape(Printer.Handle, QueryEscSupport, sizeof(Integer),
                 @QEsc, NIL)>0 then
          Escape(Printer.Handle, GetPrintingOffset, 0, NIL, @UpperLeftPagePoint)
        else
          with UpperLeftPagePoint do
          begin
            x := 0; y := 0; { assume there is no physical offset ?! }
          end;

        { now that we know the TOP and LEFT offset we finally can
          compute the BOTTOM and RIGHT offset: }
        with LowerRightPagePoint do
        begin
          x := x - UpperLeftPagePoint.x;
          { we don't want to have negative values}
          if x < 0 then
            x := 0; { assume no right printing offset }

          y := y - UpperLeftPagePoint.x;
          { we don't want to have negative values}
          if y < 0 then
            y := 0; { assume no bottom printing offset }
        end;
        { which results in LowerRightPoint containing the BOTTOM
          and RIGHT unprintable
          area offset; using these we modify the (logical, true)
          borders...}

        MLeftPrn := trunc(FPrintMarginLeft/2.54 * P1);
        MLeftPrn := MLeftPrn - UpperLeftPagePoint.x; { subtract physical offset }
        MLeft := MulDiv(MLeftPrn, P2, P1);

        MRightPrn := trunc(FPrintMarginRight/2.54 * P1);
        MRightPrn := MRightPrn - LowerRightPagePoint.x; { subtract physical offset }

        WPrn := PageWidth - (MLeftPrn + MRightPrn);
        W := MulDiv(WPrn, P2, P1);

        Curs := 0;
        PrintList.DoLogic(Canvas, 0, W, Dummy, Curs);

        MTopPrn := trunc(FPrintMarginTop/2.54 * P3);
        MTopPrn := MTopPrn - UpperLeftPagePoint.y; { subtract physical offset }

        MBottomPrn := trunc(FPrintMarginBottom/2.54 * P3);
        MBottomPrn := MBottomPrn - LowerRightPagePoint.y; { subtract physical offset }

        TopPixelsPrn := MTopPrn - UpperLeftPagePoint.y;  { subtract physical offset }
        TopPixels := MulDiv(TopPixelsPrn, P2, P3);

        HPrn := PageHeight-(MTopPrn+MBottomPrn);
        H := MulDiv(HPrn, P2, P3);    {scaled pageHeight}

        Done := False;
        HTop := 0;
        ScaledPgHt := MulDiv(PageHeight, P2, P3);
        ScaledPgWid := MulDiv(PageWidth, P2, P3);
        hrgnClip := CreateRectRgn(MLeftPrn-1, TopPixelsPrn-1, WPrn + MLeftPrn+2,
                    TopPixelsPrn + HPrn + 2);
        ARect := Rect(MLeft, TopPixels, W + MLeft, TopPixels + H);
        while (FPage <= ToPage) and not Done do
          begin
          PrintList.SetYOffset(HTop-TopPixels);
          HBot := HTop + H;
          SetMapMode(DC, mm_AnIsotropic);
          SetWindowExtEx(DC, P2, P2, Nil);
          SetViewPortExtEx(DC, P1,P3, Nil);
          SetWindowOrgEx(DC, 0, 0, Nil);
          SelectClipRgn(DC, hrgnClip);

          if FPage >= FromPage then
            begin
            PrintList.Draw(Canvas, ARect, W, MLeft, 0);
            if Assigned(FOnPrintHeader) or Assigned(FOnPrintFooter) then
              begin
              { preserve current settings of the Canvas, in case user
                  would make changes and not restore them back }
              savedFont.Assign (Canvas.Font);
              savedPen.Assign (Canvas.Pen);
              savedBrush.Assign (Canvas.Brush);
              SelectClipRgn(DC, 0);
              Align := SetTextAlign(DC, TA_Top or TA_Left or TA_NOUPDATECP);
              if Assigned(FOnPrintHeader) then
                begin
                SetWindowOrgEx(DC, 0, 0, Nil);
                FOnPrintHeader(Self, Canvas, FPage, ScaledPgWid, TopPixels, Done);
                end;
              if Assigned(FOnPrintFooter) then
                begin
                SetWindowOrgEx(DC, 0, -(TopPixels+H), Nil);
                FOnPrintFooter(Self, Canvas, FPage, ScaledPgWid,
                    ScaledPgHt-(TopPixels+H), Done);
                end;
              { restore initial Canvas settings }
              Canvas.Font.Assign(savedFont);
              Canvas.Pen.Assign(savedPen);
              Canvas.Brush.Assign(savedBrush);
              SetTextAlign(DC, Align);
              end;
            end
          else PrintList.Draw(Canvas, ARect, W, MLeft+3*W, 0); {off page}

          HTop := PrintList.PageBottom;
          if PrintList.FindLineAtPosition(HTop, LineTop, Index) = Nil then
            Done := True;
          if not Done and (FPage >= FromPage) and (FPage < ToPage) then
            NewPage;
          Inc(FPage);
          end;
        end;
    finally
      if (FromPage > FPage) then
        Printer.Abort;
      Printer.EndDoc;
      Dec(FPage);
      if hRgnClip <> 0 then DeleteObject(hrgnClip);
      Printer.Canvas.Font.Assign(SaveFont);  {restore font for others}
      SaveSize := Printer.Canvas.Font.Size;
      Printer.Canvas.Font.PixelsPerInch := P3;
      Printer.Canvas.Font.Size := SaveSize;
      SaveFont.Free;
      Printer.Title := '';
      end;
  finally
    savedFont.Free;
    savedPen.Free;
    savedBrush.Free;
    end;
finally
  PrintList.Free;
  end;
end;

{$else)   {the Delphi 2 version}

procedure ThtmlViewer.Print(FromPage, ToPage: integer);
var
  ARect: TRect;
  PrintList: TSectionList;
  P1, P2, P3, W, H, HTop, HBot, Index, Dummy: integer;
  LineTop, Curs: LongInt;
  Done: boolean;
  DC : HDC;
  vwP, OldPrinter: TvwPrinter;
  QEsc: integer;

  UpperLeftPagePoint, { these will contain Top/Left and Bottom/Right
                        unprintable area}
  LowerRightPagePoint: TPoint;
  MLeft: integer;
  MLeftPrn: integer;
  MRightPrn: integer;
  MTopPrn: integer;
  MBottomPrn: integer;
  TopPixels, TopPixelsPrn, HPrn, WPrn: integer;
  hrgnClip: THandle;
  savedFont : TFont ;
  savedPen : TPen ;
  savedBrush : TBrush ;
  Align, ScaledPgHt, ScaledPgWid: integer;

begin
FPage := 0;    
if FProcessing or (SectionList.Count = 0) then Exit;
PrintList := TSectionList.CreateCopy(SectionList);
PrintList.SetYOffset(0);
try
  savedFont := TFont.Create ;
  savedPen := TPen.Create ;
  savedBrush := TBrush.Create ;
  try
    PrintList.Printing := True;
    PrintList.SetBackground(clWhite);
    vwP := TvwPrinter.Create;
    OldPrinter := vwSetPrinter(vwP);
    FPage := 1;
    hrgnClip := 0;
    try
      with vwP do
        begin
        if (DocumentTitle <> '') then   
            vwP.Title := DocumentTitle ;
        BeginDoc;
        DC := Canvas.Handle;
        P3 := GetDeviceCaps(DC, LOGPIXELSY);
        P2 := Screen.PixelsPerInch;
        SetMapMode(DC, mm_AnIsotropic);
        P1 := GetDeviceCaps(DC, LOGPIXELSX);
        SetWindowExtEx(DC, P2, P2, Nil);
        SetViewPortExtEx(DC, P1,P3, Nil);


        { calculate the amount of space that is non-printable }

        { get PHYSICAL page width }
        QEsc := GetPhysPageSize;
        if Escape(Printer.Handle, QueryEscSupport, sizeof(Integer),
                                  @QEsc, NIL)>0 then
          Escape(Printer.Handle, GetPhysPageSize, 0, NIL, @LowerRightPagePoint)
        else { command NOT supported !}
          with LowerRightPagePoint do
          begin
            x := -1; y := -1; { signal that these values are not good at all...}
          end;

        { now compute a complete unprintable area rectangle (composed of 2*width,
          2*height) in pixels...}
        with LowerRightPagePoint do
        begin
          y := y - GetDeviceCaps(DC, VertRes);
          x := x - GetDeviceCaps(DC, HorzRes);
        end;

        { get upper left physical offset for the printer... ->
          printable area <> paper size }
        QEsc := GetPrintingOffset;
        if Escape(Printer.Handle, QueryEscSupport, sizeof(Integer),
                 @QEsc, NIL)>0 then
          Escape(Printer.Handle, GetPrintingOffset, 0, NIL, @UpperLeftPagePoint)
        else
          with UpperLeftPagePoint do
          begin
            x := 0; y := 0; { assume there is no physical offset ?! }
          end;

        { now that we know the TOP and LEFT offset we finally can
          compute the BOTTOM and RIGHT offset: }
        with LowerRightPagePoint do
        begin
          x := x - UpperLeftPagePoint.x;
          { we don't want to have negative values}
          if x < 0 then
            x := 0; { assume no right printing offset }

          y := y - UpperLeftPagePoint.x;
          { we don't want to have negative values}
          if y < 0 then
            y := 0; { assume no bottom printing offset }
        end;
        { which results in LowerRightPoint containing the BOTTOM
          and RIGHT unprintable
          area offset; using these we modify the (logical, true)
          borders...}

        MLeftPrn := trunc(FPrintMarginLeft/2.54 * P1);
        MLeftPrn := MLeftPrn - UpperLeftPagePoint.x; { subtract physical offset }
        MLeft := MulDiv(MLeftPrn, P2, P1);

        MRightPrn := trunc(FPrintMarginRight/2.54 * P1);
        MRightPrn := MRightPrn - LowerRightPagePoint.x; { subtract physical offset }

        WPrn := PageWidth - (MLeftPrn + MRightPrn);

        W := MulDiv(WPrn, P2, P1) - 2*FMarginWidth;
        Curs := 0;
        PrintList.DoLogic(Canvas, 0, W, Dummy, Curs);

        MTopPrn := trunc(FPrintMarginTop/2.54 * P3);
        MTopPrn := MTopPrn - UpperLeftPagePoint.y; { subtract physical offset }

        MBottomPrn := trunc(FPrintMarginBottom/2.54 * P3);
        MBottomPrn := MBottomPrn - LowerRightPagePoint.y; { subtract physical offset }

        TopPixelsPrn := MTopPrn - UpperLeftPagePoint.y;  { subtract physical offset }
        TopPixels := MulDiv(TopPixelsPrn, P2, P3);

        HPrn := PageHeight-(MTopPrn+MBottomPrn);
        H := MulDiv(HPrn, P2, P3);  {scaled pageHeight}
        Done := False;
        HTop := 0;
        ScaledPgHt := MulDiv(PageHeight, P2, P3);
        ScaledPgWid := MulDiv(PageWidth, P2, P3);
        hrgnClip := CreateRectRgn(MLeftPrn-1, TopPixelsPrn-1, WPrn + MLeftPrn+2,
                  TopPixelsPrn + HPrn+2);
        while (FPage <= ToPage) and not Done do
          begin
          HBot := HTop + H;
          ARect := Rect(MLeft, HTop, W + MLeft, HBot);
          SetMapMode(DC, mm_AnIsotropic);
          SetWindowExtEx(DC, P2, P2, Nil);
          SetViewPortExtEx(DC, P1,P3, Nil);
          SetWindowOrgEx(DC, 0, HTop-TopPixels, Nil);
          SelectClipRgn(DC, hrgnClip);

          if FPage >= FromPage then
            begin
            PrintList.Draw(Canvas, ARect, W, MLeft, 0);
            if Assigned(FOnPrintHeader) or Assigned(FOnPrintFooter) then
              begin
              { preserve current settings of the Canvas, in case user
                  would make changes and not restore them back }
              savedFont.Assign (Canvas.Font);
              savedPen.Assign (Canvas.Pen);
              savedBrush.Assign (Canvas.Brush);
              SelectClipRgn(DC, 0);
              Align := SetTextAlign(DC, TA_Top or TA_Left or TA_NOUPDATECP);
              if Assigned(FOnPrintHeader) then
                begin
                SetWindowOrgEx(DC, 0, 0, Nil);
                FOnPrintHeader(Self, Canvas, FPage, ScaledPgWid, TopPixels, Done);
                end;
              if Assigned(FOnPrintFooter) then
                begin
                SetWindowOrgEx(DC, 0, -(TopPixels+H), Nil);
                FOnPrintFooter(Self, Canvas, FPage, ScaledPgWid,
                    ScaledPgHt-(TopPixels+H), Done);
                end;
              { restore initial Canvas settings }
              Canvas.Font.Assign(savedFont);
              Canvas.Pen.Assign(savedPen);
              Canvas.Brush.Assign(savedBrush);
              SetTextAlign(DC, Align);
              end;
            end
          else PrintList.Draw(Canvas, ARect, W, MLeft+3*W, 0);  {off page}
          HTop := PrintList.PageBottom;
          if PrintList.FindLineAtPosition(HTop, LineTop, Index) = Nil then
            Done := True;
          if not Done and (FPage >= FromPage) and (FPage < ToPage) then
            NewPage;
          Inc(FPage);
          end;
        end;
    finally
      if hRgnClip <> 0 then DeleteObject(hrgnClip);
      if (FromPage > FPage) then
        vwPrinter.Abort
      else
        vwPrinter.EndDoc;
      Dec(FPage);
      vwSetPrinter(OldPrinter);
      vwP.Free;
      end;
  finally
    savedFont.Free ;
    savedPen.Free ;
    savedBrush.Free ;
    end;
finally
  PrintList.Free;
  end;
end;
{$endif}

function ThtmlViewer.NumPrinterPages: integer;
begin
Print(9998, 9999);
Result := FPage;
end;

procedure ThtmlViewer.BackgroundChange(Sender: TObject);
begin
PaintPanel.Color := (Sender as TSectionList).Background or $2000000;
end;

procedure ThtmlViewer.SetOnBitmapRequest(Handler: TGetBitmapEvent);
begin
FOnBitmapRequest := Handler;
SectionList.GetBitmap := Handler;
end;

procedure ThtmlViewer.SetOnImageRequest(Handler: TGetImageEvent);
begin
FOnImageRequest := Handler;
SectionList.GetImage := Handler;
end;

procedure ThtmlViewer.SetOnScript(Handler: TScriptEvent);
begin
FOnScript := Handler;
SectionList.ScriptEvent := Handler;
end;

procedure ThtmlViewer.SetOnFormSubmit(Handler: TFormSubmitEvent);
begin
FOnFormSubmit := Handler;
if Assigned(Handler) then
  SectionList.SubmitForm := SubmitForm
else SectionList.SubmitForm := Nil;
end;

procedure ThtmlViewer.SubmitForm(Sender: TObject; Action, Method: string;
                    Results: TStringList);
begin
if Assigned(FOnFormSubmit) then
  begin
  FAction := NewStr(Action);
  FMethod := NewStr(Method);
  FStringList := Results;
  PostMessage(Handle, wm_FormSubmit, 0, 0);
  end;
end;

procedure ThtmlViewer.WMFormSubmit(var Message: TMessage);
begin
FOnFormSubmit(Self, FAction^, FMethod^, FStringList);
DisposeStr(FAction);
DisposeStr(FMethod);
end;     {user disposes of the TStringList}

procedure ThtmlViewer.WMDrawBorder(var Message: TMessage);
begin
DrawBorder;
end;

function ThtmlViewer.Find(const S: String; MatchCase: boolean): boolean;
var
  ChArray: array[0..256] of char;
  Curs: LongInt;
  X: integer;
  Y, Pos: LongInt;
begin
Result := False;
if S = '' then Exit;
StrPCopy(ChArray, S);
with SectionList do
  begin
  Curs := FindString(CaretPos, ChArray, MatchCase);
  if Curs >= 0 then
    begin
    Result := True;
    SelB := Curs;
    SelE := Curs+Length(S);
    CaretPos := SelE;
    if CursorToXY(PaintPanel.Canvas, Curs, X, Y) then
      begin
      Pos := VScrollBarPosition * LongInt(VScale);
      if (Y < Pos) or
             (Y > Pos +ClientHeight-20) then    
        VScrollBarPosition := (Y - ClientHeight div 2) div VScale;   
      Invalidate;
      end;
    end;
  end;
end;

procedure ThtmlViewer.SelectAll;
begin
with SectionList do
  if (Count > 0) and not FNoSelect then
    begin
    SelB := 0;
    with TSectionBase(Items[Count-1]) do
      SelE := StartCurs + Len;
    Invalidate;
    end;
end;

procedure ThtmlViewer.Clear;
{Note: because of Frames do not clear history list here}
begin
if FProcessing then Exit;
SectionList.Clear;
SectionList.SetFonts(FFontName^, FPreFontName^, FFontSize, FFontColor,
                                  FHotSpotColor, FBackground);
FNameList.Clear;
DisposeStr(FBase); FBase := NullStr;
AssignStr(FBaseEx, '');      
DisposeStr(FBaseTarget); FBaseTarget := NullStr;
DisposeStr(FTitle); FTitle := NullStr;
VScrollBar.Max := 0;
VScrollBar.Visible := False;
VScrollBar.Height := PaintPanel.Height;
HScrollBar.Visible := False;
CaretPos := 0;
Sel1 := -1;
Invalidate;
end;

procedure ThtmlViewer.PaintWindow(DC: HDC);
begin
PaintPanel.RePaint;
end;

procedure ThtmlViewer.CopyToClipboard;
begin
SectionList.CopyToClipboard;
end;

function ThtmlViewer.GetSelTextBuf(Buffer: PChar; BufSize: LongInt): LongInt;
begin
if BufSize <= 0 then Result := 0
else Result := SectionList.GetSelTextBuf(Buffer, BufSize);
end;

{$ifdef Windows} {Delphi 1}
function ThtmlViewer.GetSelText: string;
{only reads the first 255 chars of selected text}
var
  Buf: array[0..255] of char;
  Len: integer;
begin
Len := SectionList.GetSelTextBuf(@Buf, 256);
Dec(Len);     {the #0}
if Len > 0 then
  begin
  Move(Buf, Result[1], Len);
  Result[0] := chr(Len);
  end
else Result := '';
end;

{$else}  {Delphi 2,3}
function ThtmlViewer.GetSelText: string;
var
  Len: LongInt;
begin
Len := GetSelLength;
if Len > 0 then
  begin
  SetString(Result, Nil, Len);
  SectionList.GetSelTextBuf(Pointer(Result), Len+1);
  end
else Result := '';
end;
{$endif}

function ThtmlViewer.GetSelLength: LongInt;
begin
Result := SectionList.GetSelLength;
end;

procedure ThtmlViewer.SetSelLength(Value: LongInt);
begin
if Value = 0 then
  with SectionList do
    begin
    SelE := SelB;
    Invalidate;
    end;
end;

procedure ThtmlViewer.SetNoSelect(Value: boolean);
begin
if Value <> FNoSelect then
  begin
  FNoSelect := Value;
  if Value = True then
    begin
    SectionList.SelB := -1;
    SectionList.SelE := -1;
    RePaint;
    end;
  end;
end;

procedure ThtmlViewer.UpdateImageCache;
begin
BitmapList.BumpAndCheck;
end;

procedure ThtmlViewer.SetImageCacheCount(Value: integer);
begin
Value := IntMax(0, Value);
Value := IntMin(20, Value);
if Value <> FImageCacheCount then
  begin
  FImageCacheCount := Value;
  BitmapList.SetCacheCount(FImageCacheCount);
  end;
end;

procedure ThtmlViewer.DrawBorder;
begin
if (Focused and (FBorderStyle = htFocused)) or (FBorderStyle = htSingle)
       or (csDesigning in ComponentState) then
  begin
  PaintBox.Canvas.Pen.color := clBtnText;
  end
else
  begin
  PaintBox.Canvas.Pen.Color := PaintPanel.Color or $2000000;
  end;
if Showing then
  PaintBox.Canvas.Rectangle(0,0,Width, Height);
end;

procedure ThtmlViewer.DoEnter;
begin
inherited DoEnter;
DrawBorder;
end;

procedure ThtmlViewer.DoExit;
begin
inherited DoExit;
DrawBorder;
end;

procedure ThtmlViewer.SetScrollBars(Value: TScrollStyle);
begin
if (Value <> FScrollBars) then
  begin
  FScrollBars := Value;
  if not (csLoading in ComponentState) and HandleAllocated then
    begin
    SetProcessing(True);
    try
      DoLogic;
    finally
      SetProcessing(False);
      end;
    Invalidate;
    end;
  end;
end;

{----------------ThtmlViewer.Reload}
procedure ThtmlViewer.Reload;     {reload the last file}
var
  Pos: LongInt;
begin
if FCurrentFile^ <> '' then
  begin
  Pos := Position;
  if FCurrentFileType = HTMLType then
    LoadFromFile(FCurrentFile^)
  else if FCurrentFileType = TextType then
    LoadTextFile(FCurrentFile^)
  else LoadImageFile(FCurrentFile^);
  Position := Pos;
  end;
end;

{----------------ThtmlViewer.GetOurPalette:}
function ThtmlViewer.GetOurPalette: HPalette;
begin
if ColorBits = 8 then
  Result := CopyPalette(ThePalette)
else Result := 0;
end;

{----------------ThtmlViewer.SetOurPalette}
procedure ThtmlViewer.SetOurPalette(Value: HPalette);
var
  NewPalette: HPalette;
begin
if (Value <> 0) and (ColorBits = 8) then
  begin
  NewPalette := CopyPalette(Value);
  if NewPalette <> 0 then
    begin
    if ThePalette <> 0 then
      DeleteObject(ThePalette);
    ThePalette := NewPalette;
    if FDither then SetGlobalPalette(ThePalette);
    end;
  end;
end;

{----------------ThtmlViewer.SetDither}
procedure ThtmlViewer.SetDither(Value: boolean);
begin
if (Value <> FDither) and (ColorBits = 8) then
  begin
  FDither := Value;
  if Value then SetGlobalPalette(ThePalette)
  else SetGLobalPalette(0);
  end;
end;

procedure ThtmlViewer.SetCaretPos(Value: LongInt);
begin
if Value >= 0 then
  FCaretPos := Value;
end;

{----------------ThtmlViewer.SetProcessing}
procedure ThtmlViewer.SetProcessing(Value: boolean);
begin
if FProcessing <> Value then
  begin
  FProcessing := Value;
  if Assigned(FOnProcessing) then
        FOnProcessing(Self, FProcessing);
  end;
end;

{----------------THTMLViewer.SetMarginWidth}
procedure THTMLViewer.SetMarginWidth(Value: integer);
var
  OldPos: LongInt;
  OldCursor: TCursor;
begin
if (Value <> FMarginWidth) and not FProcessing and (Value >= 0) and (Value <= 50) then
  begin
  OldCursor := Screen.Cursor;
  try
    Screen.Cursor := crHourGlass;
    SetProcessing(True);
    FMarginWidth := Value;
    if SectionList.Count > 0 then
      begin
      OldPos := Position;
      DoLogic;
      Position := OldPos;
      Invalidate;
      end;
  finally
    Screen.Cursor := OldCursor;
    SetProcessing(False);
    end;
  end;
end;

{----------------THTMLViewer.SetMarginHeight}
procedure THTMLViewer.SetMarginHeight(Value: integer);
var
  OldPos: LongInt;
  OldCursor: TCursor;
begin
if (Value <> FMarginHeight) and not FProcessing and (Value >= 0) and (Value <= 50) then
  begin
  OldCursor := Screen.Cursor;
  try
    Screen.Cursor := crHourGlass;
    SetProcessing(True);
    FMarginHeight := Value;
    if SectionList.Count > 0 then
      begin
      OldPos := Position;
      DoLogic;
      Position := OldPos;
      Invalidate;
      end;
  finally
    Screen.Cursor := OldCursor;
    SetProcessing(False);
    end;
  end;
end;

procedure THTMLViewer.SetServerRoot(Value: string);
begin
Value := Trim(Value);
if (Length(Value) >= 1) and (Value[Length(Value)] = '\') then
  SetLength(Value, Length(Value)-1);
FServerRoot := Value;
end;

{----------------TPaintPanel.CreateIt}
constructor TPaintPanel.CreateIt(AOwner: TComponent; Viewer: TComponent);

begin
  inherited Create(AOwner);
  FViewer := Viewer;
end;

{----------------TPaintPanel.Paint}
procedure TPaintPanel.Paint;
var
  MemDC: HDC;
  ABitmap: HBitmap;
  ARect: TRect;
  OldPal: HPalette;
begin
if (FViewer as ThtmlViewer).DontDraw then Exit;
OldPal := 0;
Canvas.Font := Font;
Canvas.Brush.Color := Color;
ARect := Canvas.ClipRect;
Canvas2 := TCanvas.Create;   {paint on a memory DC}
try
  MemDC := CreateCompatibleDC(Canvas.Handle);
  ABitmap := 0;
  try
    with ARect do
      begin
      ABitmap := CreateCompatibleBitmap(Canvas.Handle, Right-Left, Bottom-Top);
      if ABitmap = 0 then   raise EOutOfResources.Create('Out of Resources');
      try
        SelectObject(MemDC, ABitmap);
        SetWindowOrgEx(memDC, Left, Top, Nil);
        Canvas2.Handle := MemDC;
        DoBackground(Canvas2);
        if Assigned(FOnPaint) then FOnPaint(Self);
        OldPal := SelectPalette(Canvas.Handle, ThePalette, False);
        RealizePalette(Canvas.Handle);
        BitBlt(Canvas.Handle, Left, Top, Right-Left, Bottom-Top,
                              MemDC, Left, Top, SrcCopy);
        PostMessage(ThtmlViewer(FViewer).Handle, wm_DrawBorder, 0, 0);
      finally
        if OldPal <> 0 then SelectPalette(MemDC, OldPal, False);
        Canvas2.Handle := 0;
        end;
      end;
  finally
    DeleteDC(MemDC);
    DeleteObject(ABitmap);
  end;
finally
  Canvas2.Free;
  end;
end;

procedure TPaintPanel.DoBackground(ACanvas: TCanvas);
var
  Bitmap: TBitmap;
  H, W, WW, HH, HPos, BW, BH, DCx, DCy: integer;   
  Pos: LongInt;
  OldBrush: HBrush;
  OldPal: HPalette;
  ARect: TRect;
  DC: HDC;
  CopyFromDC: boolean;
begin
DC := ACanvas.handle;
if DC <> 0 then
  begin
  Pos := (FViewer as ThtmlViewer).VScrollBarPosition * LongInt(VScale);
  ARect := Canvas.ClipRect;

  OldPal := SelectPalette(DC, ThePalette, False);
  RealizePalette(DC);
  ACanvas.Brush.Color := Color or $2000000;
  OldBrush := SelectObject(DC, ACanvas.Brush.Handle);
  try
    with FViewer as ThtmlViewer do
      begin
      if True then   {note:  this code for later use for Watermarks}
        HPos := HScrollBar.Position
      else
        begin
        HPos := 0;
        Pos := 0;
        end;
      Bitmap := SectionList.BackgroundBitmap;
      if SectionList.ShowImages and not SectionList.Printing
                                 and Assigned(Bitmap) then
        try
          DCx := 0;  DCy := 0;
          BW := Bitmap.Width;
          BH := Bitmap.Height;
          HH := (Pos div BH) * BH - Pos;
          WW := (HPos div BW) * BW - HPos;
          while HH < ARect.Top do
            Inc(HH, BH);
          while WW < ARect.Left do
            Inc(WW, BW);
          {to use the fast method, the bitmap must be drawn entirely within the
           viewable area}
          if (HH+BH <= ARect.Bottom) and ((WW+BW <= ARect.Right)
                  or ((WW = ARect.Left) and (BW >= ARect.Right-ARect.Left))) then
            begin
            CopyFromDC := True;   {fast}
            DCx := WW;
            DCy := HH;
            end
          else CopyFromDC := False;
          Dec(HH, BH);
          Dec(WW, BW);
          H := HH;
          if CopyFromDC then
            BitBlt(DC, DCx, DCy, BW, BH, Bitmap.Canvas.Handle, 0, 0, SRCCOPY);  
          repeat
            W := WW;
            repeat
              if CopyFromDC then
                BitBlt(DC, W, H, BW, BH, DC, DCx, DCy, SRCCOPY)   
              else
                BitBlt(DC, W, H, BW, BH, Bitmap.Canvas.Handle, 0, 0, SRCCOPY);   
              Inc(W, BW);
            until W >= ARect.Right;
            Inc(H, BH);
          until H >= ARect.Bottom;
        except
          ACanvas.FillRect(ARect);
        end
      else
        begin
        ACanvas.FillRect(ARect);
        end;
      end;
  finally
    SelectObject(DC, OldBrush);
    SelectPalette(DC, OldPal, False);
    RealizePalette(DC);
    end;
  end;

end;

procedure TPaintPanel.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
{Only do background now if resizing the window}
if Sizing and (Message.DC <> 0) then
  begin
  Canvas.Handle := Message.DC;
  try
    DoBackground(Canvas);
  finally
    Canvas.Handle := 0;
    Sizing := False;
    end;
  end;
Message.Result := 1;   {it's erased}
end;

{----------------TPaintPanel.WMLButtonDblClk} 
procedure TPaintPanel.WMLButtonDblClk(var Message: TWMMouse);
begin
if Message.Keys and MK_LButton <> 0 then
  ThtmlViewer(FViewer).HTMLMouseDblClk(Message.XPos, Message.YPos);
end;

{$ifdef Win32}
{----------------T32ScrollBar.SetParams}
procedure T32ScrollBar.SetParams(APosition, APage, AMin, AMax: Integer);
var
  ScrollInfo: TScrollInfo;
begin
if (APosition <> FPosition) or (APage <> FPage) or (AMin <> FMin)
              or (AMax <> FMax) then
  with ScrollInfo do
    begin
    cbSize := SizeOf(ScrollInfo);
    fMask := SIF_ALL;
    nPos := APosition;
    nPage := APage;
    nMin := AMin;
    nMax := AMax;
    SetScrollInfo(Handle, SB_CTL, ScrollInfo, True);
    FPosition := APosition;
    FPage := APage;
    FMin := AMin;
    FMax := AMax;
    end;
end;

procedure T32ScrollBar.SetPosition(Value: integer);
begin
SetParams(Value, FPage, FMin, FMax);
end;

procedure T32ScrollBar.SetMin(Value: Integer);
begin
  SetParams(FPosition, FPage, Value, FMax);
end;

procedure T32ScrollBar.SetMax(Value: Integer);
begin
  SetParams(FPosition, FPage, FMin, Value);
end;

procedure T32ScrollBar.CNVScroll(var Message: TWMVScroll);
var
  SPos: LongInt;
  ScrollInfo: TScrollInfo;
  OrigPos: LongInt;
  TheChange: LongInt;
begin
Parent.SetFocus;
with ThtmlViewer(Parent) do
  begin
  ScrollInfo.cbSize := SizeOf(ScrollInfo);
  ScrollInfo.fMask := SIF_ALL;
  GetScrollInfo(Self.Handle, SB_CTL, ScrollInfo);
  if TScrollCode(Message.ScrollCode) = scTrack then
    begin
    SPos := ScrollInfo.nTrackPos;
    OrigPos := SPos;
    end
  else
    begin
    SPos := ScrollInfo.nPos;
    OrigPos := SPos;
    case TScrollCode(Message.ScrollCode) of
      scLineUp:
        Dec(SPos, SmallChange);
      scLineDown:
        Inc(SPos, SmallChange);
      scPageUp:
        Dec(SPos, LargeChange);
      scPageDown:
        Inc(SPos, LargeChange);
      scTop:
        SPos := 0;
      scBottom:
        SPos := (MaxVertical - PaintPanel.Height) div VScale;
      end;
    end;
  if SPos < 0 then SPos := 0;
  SPos := IntMin(SPos, (MaxVertical - PaintPanel.Height) div VScale);

  Self.SetPosition(SPos);

  SectionList.SetYOffset(SPos * VScale);
  TheChange := OrigPos-SPos;

  if Abs(TheChange)=SmallChange then begin
         ScrollWindow(PaintPanel.Handle,0,TheChange,NIL,NIL);
         PaintPanel.Update;
     end else begin
        PaintPanel.Invalidate;
     end;
  end;
end;
{$endif}

function THTMLEditor.GetVerbCount: integer;     
begin
  Result := 1;
end;

function THTMLEditor.GetVerb(index: Integer): string;  
begin
  Result := 'About..';
end;

procedure THTMLEditor.ExecuteVerb(index:integer);   
begin
  MessageDlg('ThtmlViewer'+#13#13+
             'Version     : 6.01'+#13#13+
             'Copyright  : 1995-8 by L. David Baldwin, All Rights Reserved'+#13#13+
             'Support    : dbaldwin@pbear.com'+#13#13+
             'Web Site : http://www.pbear.com/ '
             ,mtInformation,[mbOk],0)
end;

procedure Register;
begin
RegisterComponents('Samples', [THTMLViewer]);
RegisterComponentEditor(THTMLViewer, THTMLEditor);     
end;

end.

